# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'galacteek/ui/addhashmarkdialog.ui'
#
# Created by: PyQt5 UI code generator 5.10
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_AddHashmarkDialog(object):
    def setupUi(self, AddHashmarkDialog):
        AddHashmarkDialog.setObjectName("AddHashmarkDialog")
        AddHashmarkDialog.resize(1065, 418)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(AddHashmarkDialog.sizePolicy().hasHeightForWidth())
        AddHashmarkDialog.setSizePolicy(sizePolicy)
        AddHashmarkDialog.setMaximumSize(QtCore.QSize(16777215, 16777215))
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/share/icons/hashmarks.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        AddHashmarkDialog.setWindowIcon(icon)
        AddHashmarkDialog.setStyleSheet("")
        self.verticalLayout = QtWidgets.QVBoxLayout(AddHashmarkDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setLabelAlignment(QtCore.Qt.AlignCenter)
        self.formLayout.setFormAlignment(QtCore.Qt.AlignHCenter|QtCore.Qt.AlignTop)
        self.formLayout.setVerticalSpacing(10)
        self.formLayout.setObjectName("formLayout")
        self.hashmarksIconLabel = QtWidgets.QLabel(AddHashmarkDialog)
        self.hashmarksIconLabel.setText("")
        self.hashmarksIconLabel.setObjectName("hashmarksIconLabel")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.hashmarksIconLabel)
        self.resourceLabel = QtWidgets.QLabel(AddHashmarkDialog)
        self.resourceLabel.setMinimumSize(QtCore.QSize(800, 0))
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.resourceLabel.setFont(font)
        self.resourceLabel.setText("")
        self.resourceLabel.setObjectName("resourceLabel")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.resourceLabel)
        self.label = QtWidgets.QLabel(AddHashmarkDialog)
        self.label.setObjectName("label")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label)
        self.title = QtWidgets.QLineEdit(AddHashmarkDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.title.sizePolicy().hasHeightForWidth())
        self.title.setSizePolicy(sizePolicy)
        self.title.setMaximumSize(QtCore.QSize(500, 16777215))
        self.title.setMaxLength(128)
        self.title.setObjectName("title")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.title)
        self.label_7 = QtWidgets.QLabel(AddHashmarkDialog)
        self.label_7.setObjectName("label_7")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_7)
        self.description = QtWidgets.QTextEdit(AddHashmarkDialog)
        self.description.setMaximumSize(QtCore.QSize(600, 16777215))
        self.description.setAcceptRichText(False)
        self.description.setObjectName("description")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.description)
        self.label_6 = QtWidgets.QLabel(AddHashmarkDialog)
        self.label_6.setObjectName("label_6")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.label_6)
        self.comment = QtWidgets.QLineEdit(AddHashmarkDialog)
        self.comment.setMaximumSize(QtCore.QSize(600, 16777215))
        self.comment.setMaxLength(128)
        self.comment.setObjectName("comment")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.comment)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.formLayout.setItem(4, QtWidgets.QFormLayout.FieldRole, spacerItem)
        self.label_2 = QtWidgets.QLabel(AddHashmarkDialog)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(5, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.category = QtWidgets.QComboBox(AddHashmarkDialog)
        self.category.setMaximumSize(QtCore.QSize(600, 16777215))
        self.category.setObjectName("category")
        self.formLayout.setWidget(5, QtWidgets.QFormLayout.FieldRole, self.category)
        self.label_4 = QtWidgets.QLabel(AddHashmarkDialog)
        self.label_4.setObjectName("label_4")
        self.formLayout.setWidget(6, QtWidgets.QFormLayout.LabelRole, self.label_4)
        self.newCategory = QtWidgets.QLineEdit(AddHashmarkDialog)
        self.newCategory.setMaximumSize(QtCore.QSize(600, 16777215))
        self.newCategory.setMaxLength(128)
        self.newCategory.setObjectName("newCategory")
        self.formLayout.setWidget(6, QtWidgets.QFormLayout.FieldRole, self.newCategory)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.formLayout.setItem(7, QtWidgets.QFormLayout.FieldRole, spacerItem1)
        self.label_8 = QtWidgets.QLabel(AddHashmarkDialog)
        self.label_8.setObjectName("label_8")
        self.formLayout.setWidget(9, QtWidgets.QFormLayout.LabelRole, self.label_8)
        self.pinCombo = QtWidgets.QComboBox(AddHashmarkDialog)
        self.pinCombo.setMaximumSize(QtCore.QSize(250, 16777215))
        self.pinCombo.setObjectName("pinCombo")
        self.formLayout.setWidget(9, QtWidgets.QFormLayout.FieldRole, self.pinCombo)
        self.label_3 = QtWidgets.QLabel(AddHashmarkDialog)
        self.label_3.setText("")
        self.label_3.setObjectName("label_3")
        self.formLayout.setWidget(10, QtWidgets.QFormLayout.LabelRole, self.label_3)
        self.share = QtWidgets.QCheckBox(AddHashmarkDialog)
        self.share.setObjectName("share")
        self.formLayout.setWidget(10, QtWidgets.QFormLayout.FieldRole, self.share)
        self.verticalLayout.addLayout(self.formLayout)
        self.buttonBox = QtWidgets.QDialogButtonBox(AddHashmarkDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(AddHashmarkDialog)
        self.buttonBox.accepted.connect(AddHashmarkDialog.accept)
        self.buttonBox.rejected.connect(AddHashmarkDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(AddHashmarkDialog)

    def retranslateUi(self, AddHashmarkDialog):
        _translate = QtCore.QCoreApplication.translate
        AddHashmarkDialog.setWindowTitle(_translate("AddHashmarkDialog", "Add new hashmark"))
        self.label.setText(_translate("AddHashmarkDialog", "Title"))
        self.label_7.setText(_translate("AddHashmarkDialog", "Description"))
        self.label_6.setText(_translate("AddHashmarkDialog", "Comment"))
        self.label_2.setText(_translate("AddHashmarkDialog", "Category"))
        self.label_4.setText(_translate("AddHashmarkDialog", "or create new category, e.g news/world"))
        self.label_8.setText(_translate("AddHashmarkDialog", "Pinning"))
        self.share.setText(_translate("AddHashmarkDialog", "Share (public)"))

from . import galacteek_rc
